(** Client configuration *)
module Config : sig
  open Okcrypto

  type 's t

  (**
    [make storage] creates new configuration using given session storage.

    - [~base_url] is a base external URL of the application.
    - [~app_name] is an application name that is displayed to users.
    - [~public_encryption_key], [~secret_encryption_key], [~public_signing_key]
      and [~secret_signing_key] are key pairs generated for the application to
      encrypt and sign requests.
    - [~session_duration] controls how long generated session cookie will be
      valid (default: 14 days)
    - [~confirmation_duration] controls how long generated authentication
      confirmation will be valid (default: 10 minutes)
  *)
  val make :
    ?session_duration:Core_kernel.Time.Span.t ->
    ?confirmation_duration:Core_kernel.Time.Span.t ->
    base_url:string ->
    app_name:string ->
    public_encryption_key:Public_key.t ->
    secret_encryption_key:Secret_key.t ->
    public_signing_key:Public_key.t ->
    secret_signing_key:Secret_key.t ->
    's ->
    's t

  (** Returns a base external URL of the application *)
  val get_base_url : _ t -> string

  (** Returns application name *)
  val get_app_name : _ t -> string

  (** Returns public encryption key *)
  val get_public_encryption_key : _ t -> Public_key.t

  (** Returns secret encryption key *)
  val get_secret_encryption_key : _ t -> Secret_key.t

  (** Returns public signing key *)
  val get_public_signing_key : _ t -> Public_key.t

  (** Returns secret signing key *)
  val get_secret_signing_key : _ t -> Secret_key.t

  (** Returns session duration *)
  val get_session_duration : _ t -> Core_kernel.Time.Span.t

  (** Returns confirmation duration *)
  val get_confirmation_duration : _ t -> Core_kernel.Time.Span.t

  (** Returns session storage *)
  val get_storage : 's t -> 's
end

(** Signature of a session storage module *)
module type Storage = sig
  type t

  (** Returns value of the given [~key] if it exists in the storage *)
  val find : t -> key:string -> string option

  (** Sets the value of the given [~key] to [~value] *)
  val set : t -> key:string -> value:string -> unit

  (** Removes the value of the given [~key] *)
  val remove : t -> key:string -> unit
end

(** Signature of a client module *)
module type S = sig
  type storage

  (** Middleware for user authentication via Okauth. *)
  val middleware : storage Config.t -> Okweb.Middleware.t
end

(** Simple in-memory storage backed by [Hashtbl] *)
module In_memory_storage : sig
  include Storage

  (** Creates new empty in-memory storage *)
  val empty : t

  (** Returns size of the storage *)
  val size : t -> int
end

[@@@warning "-67"]

(** Creates new OkAuth client module using given OkWeb HTTP client and storage modules *)
module Make (Clt : module type of Okweb.Client) (St : Storage) :
  S with type storage := St.t
