open Base
open Cmdliner

module Adm = Okauth_server.Administrator.Make (struct
  module Terminal = Okauth_server.Terminal.Real_world

  let db_path = Okauth_server.db_path
end)

let app_name_arg =
  let doc = "Display name of the app" in
  Arg.(
    required
    & opt (some string) None
    & info [ "n"; "name" ] ~docv:"APP-NAME" ~doc)

let app_encryption_key_arg =
  let doc = "Public encryption key of the app" in
  Arg.(
    required
    & opt (some string) None
    & info [ "e"; "encryption-key" ] ~docv:"APP-ENCRYPTION-KEY" ~doc)

let app_signing_key_arg =
  let doc = "Public signing key of the app" in
  Arg.(
    required
    & opt (some string) None
    & info [ "s"; "signing-key" ] ~docv:"APP-SIGNING-KEY" ~doc)

let tresult term =
  let result_to_ret = function
    | Result.Ok _ -> `Ok ()
    | Result.Error msg -> `Error (false, msg)
  in
  Term.(ret (const result_to_ret $ term))

let approve_app_cmd =
  let doc =
    "Approves the application allowing it to send authentication requests to \
     this OkAuth instance"
  in
  ( tresult
      Term.(
        const Adm.approve_app $ app_encryption_key_arg $ app_signing_key_arg
        $ app_name_arg),
    Term.info "approve-app" ~doc )

let disapprove_app_cmd =
  let doc =
    "Revokes the application's approval blocking it from authenticating users \
     of this OkAuth instance"
  in
  ( tresult Term.(const Adm.disapprove_app $ app_encryption_key_arg),
    Term.info "disapprove-app" ~doc )

let list_apps_cmd =
  let doc = "List apps approved in this OkAuth instance" in
  (tresult Term.(const Adm.list_apps $ const ()), Term.info "list-apps" ~doc)

let user_id_str =
  let parse = function
    | str when String.for_all str ~f:Char.is_alphanum && String.length str >= 3
      ->
        Ok str
    | _ ->
        Error
          (`Msg
            "should be at least 3 characters long and only have latin letters \
             and digits")
  in
  Arg.(conv (parse, conv_printer string))

let user_id_arg =
  let doc =
    "ID of a user (alphanumeric, case-sensitive, at least 3 characters long)"
  in
  Arg.(
    required
    & opt (some user_id_str) None
    & info [ "i"; "uid" ] ~docv:"USER-ID" ~doc)

let user_name_arg =
  let doc = "Display name of a user" in
  Arg.(
    required
    & opt (some string) None
    & info [ "n"; "name" ] ~docv:"USER-NAME" ~doc)

let add_user_cmd =
  let doc = "Creates new user" in
  ( tresult Term.(const Adm.add_user $ user_id_arg $ user_name_arg),
    Term.info "add-user" ~doc )

let remove_user_cmd =
  let doc = "Removes user" in
  ( tresult Term.(const Adm.remove_user $ user_id_arg),
    Term.info "remove-user" ~doc )

let change_password_cmd =
  let doc = "Changes user's password" in
  ( tresult Term.(const Adm.change_password $ user_id_arg),
    Term.info "change-password" ~doc )

let list_users_cmd =
  let doc = "List users of this OkAuth instance" in
  (tresult Term.(const Adm.list_users $ const ()), Term.info "list-users" ~doc)

let init_cmd =
  let doc = "Initializes OkAuth database" in
  (tresult Term.(const Adm.init $ const ()), Term.info "init" ~doc)

let default_cmd =
  let doc = "Set up and manage OkAuth instance" in
  ( Term.(ret (const (fun () -> `Help (`Pager, None)) $ const ())),
    Term.info "okauthadm" ~version:[%blob "../../version"] ~doc )

let () =
  let cmds =
    [
      add_user_cmd;
      approve_app_cmd;
      change_password_cmd;
      disapprove_app_cmd;
      init_cmd;
      list_apps_cmd;
      list_users_cmd;
      remove_user_cmd;
    ]
  in
  Term.(exit @@ eval_choice default_cmd cmds)
