open Base
open Oksql
open Okcrypto

(** Signature of a module implementing administrative interface of OkAuth *)
module type S = sig
  (** Command that initializes the database *)
  val init : unit -> (unit, string) Result.t

  (** Interactive command that adds a new user *)
  val add_user : string -> string -> (unit, string) Result.t

  (** Interactive command that removes an user *)
  val remove_user : string -> (unit, string) Result.t

  (** Interactive command that changes user's password *)
  val change_password : string -> (unit, string) Result.t

  (** Command that lists users *)
  val list_users : unit -> (unit, string) Result.t

  (** Command that approves an app *)
  val approve_app : string -> string -> string -> (unit, string) Result.t

  (** Interactive command that disapproves an app *)
  val disapprove_app : string -> (unit, string) Result.t

  (** Command that lists approved apps *)
  val list_apps : unit -> (unit, string) Result.t
end

(** Administrator's runtime dependencies *)
module type Params = sig
  (** Terminal implementation to use *)
  module Terminal : Terminal.S

  (** Path to the database or ":memory:" to use in-memory DB *)
  val db_path : string
end

(** Makes a new administrator instance *)
module Make (P : Params) : S = struct
  module T = P.Terminal

  let ok_done () =
    T.print_endline "Done";
    Result.Ok ()

  let ok_aborted () =
    T.print_endline "Aborted";
    Result.Ok ()

  let in_db fn =
    try
      with_db P.db_path (fun db ->
          if Migration.check db ~migrations:Schema.migrations then fn db
          else
            Result.fail
              "Database is corrupted or not initialized. Please run `okauthadm \
               init` to initialize it.")
    with
    | Sqlite_error (_, msg) -> Result.fail msg
    | Material_decoding_error key ->
        Result.fail
          (Printf.sprintf "'%s' is not a valid Base64-encoded string" key)

  let if_user_exists db ~user_id ~f =
    match User.find_by_id db ~id:user_id with
    | None -> Result.fail (Printf.sprintf "User '%s' is not found" user_id)
    | Some _ -> f ()

  let if_user_doesnt_exist db ~user_id ~f =
    match User.find_by_id db ~id:user_id with
    | None -> f ()
    | Some _ -> Result.fail (Printf.sprintf "User '%s' already exists" user_id)

  let if_app_is_approved db ~encryption_key ~f =
    match Application.find_by_encryption_key db ~encryption_key with
    | None -> Result.fail "Application is not approved yet"
    | Some _ -> f ()

  let if_app_is_not_approved db ~encryption_key ~f =
    match Application.find_by_encryption_key db ~encryption_key with
    | Some _ -> Result.fail "Application is already approved"
    | None -> f ()

  let print_table headers rows =
    let col_widths =
      Array.mapi headers ~f:(fun ncol header ->
          let rows_widths =
            Array.map rows ~f:(fun row -> String.length row.(ncol))
          in
          Int.max (String.length header)
            ( rows_widths
            |> Array.max_elt ~compare:Int.compare
            |> Option.value ~default:0 ))
    in
    let align ?(fill = ' ') ncol str =
      String.concat
        [ str; String.make (col_widths.(ncol) - String.length str) fill ]
    in
    let print_row row =
      Array.iteri row ~f:(fun ncol row ->
          let open Printf in
          match ncol with
          | 0 -> sprintf "%s " (align ncol row) |> T.print_string
          | _ -> sprintf "  %s " (align ncol row) |> T.print_string)
    in
    print_row headers;
    T.print_endline "";
    Array.iter rows ~f:(fun row ->
        print_row row;
        T.print_endline "")

  let init () =
    try
      with_db P.db_path (fun db ->
          Migration.apply db ~migrations:Schema.migrations);
      ok_done ()
    with Sqlite_error (_, msg) -> Result.fail msg

  let add_user user_id user_name =
    in_db (fun db ->
        if_user_doesnt_exist db ~user_id ~f:(fun () ->
            T.print_endline (Printf.sprintf "Adding user '%s'" user_id);
            let password = T.read_password () in
            T.print_endline
              "Generating keys and storing user in the database...";
            ignore (User.create db ~id:user_id ~name:user_name ~password);
            ok_done ()))

  let remove_user user_id =
    in_db (fun db ->
        if_user_exists db ~user_id ~f:(fun () ->
            if
              T.await_confirmation (Printf.sprintf "Removing user '%s'" user_id)
            then (
              User.remove db ~id:user_id;
              ok_done () )
            else ok_aborted ()))

  let change_password user_id =
    in_db (fun db ->
        if_user_exists db ~user_id ~f:(fun () ->
            T.print_endline
              (Printf.sprintf "Changing password for '%s'" user_id);
            let new_password = T.read_password () in
            User.change_password db ~id:user_id ~new_password;
            ok_done ()))

  let list_users () =
    in_db (fun db ->
        let rows =
          User.find_all db
          |> Array.map ~f:(fun user ->
                 [|
                   User.get_id user;
                   User.get_name user;
                   User.get_public_key user |> Public_key.to_base64;
                 |])
        in
        let headers = [| "ID"; "NAME"; "PUBLIC KEY" |] in
        print_table headers rows;
        Result.Ok ())

  let approve_app enc_key_str sign_key_str name =
    in_db (fun db ->
        let encryption_key = Public_key.of_base64_exn enc_key_str in
        let signing_key = Public_key.of_base64_exn sign_key_str in
        if_app_is_not_approved db ~encryption_key ~f:(fun () ->
            T.print_endline (Printf.sprintf "Approving app '%s'" enc_key_str);
            ignore (Application.create db ~name ~encryption_key ~signing_key);
            ok_done ()))

  let disapprove_app enc_key_str =
    in_db (fun db ->
        let encryption_key = Public_key.of_base64_exn enc_key_str in
        if_app_is_approved db ~encryption_key ~f:(fun () ->
            if
              T.await_confirmation
                (Printf.sprintf "Disapproving app '%s'" enc_key_str)
            then (
              Application.remove db ~encryption_key;
              ok_done () )
            else ok_aborted ()))

  let list_apps () =
    in_db (fun db ->
        let rows =
          Application.find_all db
          |> Array.map ~f:(fun app ->
                 [|
                   Application.get_name app;
                   app |> Application.get_encryption_key |> Public_key.to_base64;
                 |])
        in
        let headers = [| "NAME"; "PUBLIC KEY" |] in
        print_table headers rows;
        Result.Ok ())
end
