open Base
open Lwt.Infix
open Oksql
open Okcrypto

(** Signature of a module implementing daemon interface of OkAuth *)
module type S = sig
  (** Initializes the daemon *)
  val init : unit -> (unit, string) Result.t

  (** Processes request to the daemon *)
  val process_request : Okweb.Handler.t
end

(** Daemon's runtime dependencies *)
module type Params = sig
  (** Path to the database or ":memory:" to use in-memory DB *)
  val db_path : string

  (** Logs message to syslog or any other supported log storage *)
  val log : string -> unit
end

(** Makes a new daemon instance *)
module Make (P : Params) : S = struct
  open Base.Printf
  open Okweb

  let init () =
    try
      Okcrypto.init ();
      with_db ~read_only:true P.db_path (fun db ->
          if not (Migration.check db ~migrations:Schema.migrations) then
            Result.fail
              "Database is corrupted or not initialized. Please run `okauthadm \
               init` to initialize it"
          else Result.Ok ())
    with
    | Sqlite_error (_, reason) -> Result.fail reason
    | Initialization_error -> Result.fail "failed to initialize libsodium"

  let warn msg = sprintf "warn: %s" msg |> P.log

  let set_cache_forever_header resp =
    resp
    |> Response.with_header ~key:"Cache-Control"
         ~value:"public, immutable, max-age=31536000"

  let set_content_type typ resp =
    resp |> Response.with_header ~key:"Content-Type" ~value:typ

  let make_error_page ~status ~title ~message =
    let template = Mustache.of_string [%blob "web/error_page.html"] in
    let params =
      `O [ ("error_title", `String title); ("error_message", `String message) ]
    in
    Response.of_status status
    |> set_content_type "text/html; charset=utf-8"
    |> Response.with_string_body (Mustache.render template params)

  let get_index_page =
    Response.(
      ok |> set_cache_forever_header
      |> with_string_body [%blob "web/index_page.html"])

  let get_okauth_css =
    Response.(
      ok |> set_cache_forever_header
      |> set_content_type "text/css"
      |> with_string_body [%blob "web/okauth.css"])

  let get_user_public_key uid =
    with_db ~read_only:true P.db_path (fun db ->
        let key =
          match User.find_by_id db ~id:uid with
          | Some user -> User.get_public_key user
          | None ->
              sprintf "public key request for a non-existent user: %s" uid
              |> warn;
              generate_seed_keypair uid |> fst
        in
        Response.(
          ok
          |> set_content_type "text/plain"
          |> with_string_body (Public_key.to_base64 key)))

  let get_auth_page uid aepk nonce_str payload_str signature_str =
    with_db ~read_only:true P.db_path (fun db ->
        match
          ( User.find_by_id db ~id:uid,
            Application.find_by_encryption_key db ~encryption_key:aepk )
        with
        | opt_user, Some app ->
            if Option.is_none opt_user then
              sprintf "authentication request for a non-existent user: %s" uid
              |> warn;
            let template = Mustache.of_string [%blob "web/login_page.html"] in
            let params =
              `O
                [
                  ("app_name", `String (Application.get_name app));
                  ("user_id", `String uid);
                  ( "app_encryption_key",
                    `String
                      ( app |> Application.get_encryption_key
                      |> Public_key.to_base64 ) );
                  ("nonce", `String nonce_str);
                  ("payload", `String payload_str);
                  ("signature", `String signature_str);
                ]
            in
            Response.(
              ok
              |> set_content_type "text/html; charset=utf-8"
              |> with_string_body (Mustache.render template params))
        | _ ->
            sprintf "authentication request from a non-approved app: %s"
              (Public_key.to_base64 aepk)
            |> warn;
            make_error_page ~status:403 ~title:"Authentication Failed"
              ~message:
                "Application is not approved to make authentication requests.")

  let authenticate uid password aepk nonce payload signature =
    with_db ~read_only:true P.db_path (fun db ->
        match
          ( User.find_by_id db ~id:uid,
            Application.find_by_encryption_key db ~encryption_key:aepk )
        with
        | Some user, Some app -> (
            let pw_is_valid =
              verify_password ~password ~hash:(User.get_password_hash user)
            in
            let signature_is_valid =
              verify_signature
                (Cipher_text.to_string payload ^ Nonce.to_string nonce)
                signature
                (Application.get_signing_key app)
            in
            match (pw_is_valid, signature_is_valid) with
            | true, true ->
                let app_url =
                  decrypt payload
                    (Application.get_encryption_key app)
                    (User.get_secret_key user) nonce
                  |> Plain_text.to_string
                in
                Response.redirect app_url
            | false, _ ->
                sprintf "invalid password for the user: %s" uid |> warn;
                make_error_page ~status:403 ~title:"Authentication Failed"
                  ~message:"User ID or password are invalid."
            | _ ->
                sprintf
                  "invalid signature for authentication request for the user: \
                   %s"
                  uid
                |> warn;
                make_error_page ~status:400 ~title:"Verification Failed"
                  ~message:
                    "Request verification failed. Please report this error to \
                     the application requesting authentication." )
        | _, None ->
            sprintf "authentication request from a non-approved app: %s"
              (Public_key.to_base64 aepk)
            |> warn;
            make_error_page ~status:403 ~title:"Authentication Failed"
              ~message:
                "Application is not approved to make authentication requests."
        | _ ->
            sprintf "authentication request for a non-existent user: %s" uid
            |> warn;
            make_error_page ~status:403 ~title:"Authentication Failed"
              ~message:"User ID or password are invalid.")

  let do_process_request req =
    match (Request.get_http_method req, Request.get_path req) with
    | "GET", [] -> Lwt.return get_index_page
    | "GET", [ "okauth.css" ] -> Lwt.return get_okauth_css
    | "GET", [ "v1"; uid; "public_key" ] -> Lwt.return (get_user_public_key uid)
    | "GET", [ "v1"; uid; "authenticate" ] ->
        let qparam key = Request.find_query_param req ~key in
        Lwt.return
          ( match (qparam "k", qparam "n", qparam "p", qparam "s") with
          | Some aepk_str, Some nonce_str, Some payload_str, Some sign_str -> (
              try
                let aepk = Public_key.of_base64_exn aepk_str in
                get_auth_page uid aepk nonce_str payload_str sign_str
              with Material_decoding_error str ->
                make_error_page ~status:400 ~title:"Bad Request"
                  ~message:
                    (sprintf
                       "Parameter is incorrectly encoded: %s. Please report \
                        this error to the application requesting \
                        authentication."
                       str) )
          | _ ->
              make_error_page ~status:400 ~title:"Bad Request"
                ~message:
                  "Required parameters are missing. Please report this error \
                   to the application requesting authentication." )
    | "POST", [ "v1"; uid; "authenticate" ] -> (
        req |> Request.get_body |> Body.read_up_to >|= Uri.query_of_encoded
        >|= fun body ->
        let bparam name =
          body
          |> List.find_map ~f:(fun (k, v) ->
                 List.hd v
                 |> Option.some_if (String.equal k name)
                 |> Option.join)
        in
        match
          (bparam "k", bparam "n", bparam "p", bparam "s", bparam "password")
        with
        | ( Some aepk_str,
            Some nonce_str,
            Some payload_str,
            Some signature_str,
            Some password ) -> (
            try
              let aepk = Public_key.of_base64_exn aepk_str in
              let nonce = Nonce.of_base64_exn nonce_str in
              let payload = Cipher_text.of_base64_exn payload_str in
              let signature = Signature.of_base64_exn signature_str in
              authenticate uid password aepk nonce payload signature
            with Material_decoding_error str ->
              make_error_page ~status:400 ~title:"Bad Request"
                ~message:(sprintf "Parameter is incorrectly encoded: %s" str) )
        | _ ->
            make_error_page ~status:400 ~title:"Bad Request"
              ~message:"Required parameters are missing." )
    | _ ->
        Lwt.return
          (make_error_page ~status:404 ~title:"Not Found"
             ~message:"The page you're looking for is not found")

  let process_request = Looksok.middleware do_process_request
end
