open Alcotest
open Base
open Oksql
open Utils
open Okauth_server
open Okcrypto
open Printf
open Lwt.Infix

let extract_error_msg body =
  let open Soup in
  let page = parse body in
  let title = page $ "div.ok-container > h1" |> R.leaf_text in
  let msg = page $ "div.ok-container > p" |> R.leaf_text in
  (title, msg)

let process_request (module D : Daemon.S) req =
  let open Okweb in
  Lwt_main.run
    ( D.process_request req >>= fun resp ->
      resp |> Response.get_body |> Body.read_up_to >|= fun body -> (resp, body)
    )

let test_get_v1_uid_public_key =
  test_case "GET /v1/UID/public_key" `Quick (fun () ->
      with_tmp_file (fun db_path ->
          let module D = Daemon.Make (struct
            let db_path = db_path

            let log = ignore
          end) in
          let uid = "testuser" in
          let user_pk =
            with_db db_path (fun db ->
                Migration.apply db ~migrations:Schema.migrations;
                let user =
                  User.create db ~id:"testuser" ~password:"password"
                    ~name:"Test User"
                in
                user |> User.get_public_key |> Public_key.to_base64)
          in

          let open Okweb in
          let get_user_pk_req uid =
            Request.(
              get (sprintf "/v1/%s/public_key" uid) |> with_string_body "")
          in

          let invalid_uid = "invalid_uid" in
          let resp, body =
            process_request (module D) (get_user_pk_req invalid_uid)
          in
          (check int) "endpoint does not return 200" 200
            (Response.get_status resp);
          (check string) "endpoint returns invalid public key"
            (generate_seed_keypair invalid_uid |> fst |> Public_key.to_base64)
            body;

          let resp, body = process_request (module D) (get_user_pk_req uid) in
          (check int) "endpoint does not return 200" 200
            (Response.get_status resp);
          (check (option string))
            "endpoint returned invalid Content-Type" (Some "text/plain")
            (Response.find_header resp ~key:"Content-Type");
          (check string) "endpoint returned incorrect public key" user_pk body))

let test_get_v1_uid_authenticate =
  test_case "GET /v1/UID/authenticate" `Quick (fun () ->
      with_tmp_file (fun db_path ->
          let module D = Daemon.Make (struct
            let db_path = db_path

            let log = ignore
          end) in
          let uid, user_name = ("testuser", "Test User") in
          let app_name = "Test App" in
          let aepk, _ = generate_keypair Key_purpose.Encryption in
          let aspk, _ = generate_keypair Key_purpose.Signing in
          with_db db_path (fun db ->
              Migration.apply db ~migrations:Schema.migrations;
              ignore
                (User.create db ~id:uid ~password:"password" ~name:user_name);
              ignore
                (Application.create db ~encryption_key:aepk ~signing_key:aspk
                   ~name:app_name));

          let open Okweb in
          let authenticate_req uid params =
            let base_uri = sprintf "/v1/%s/authenticate" uid |> Uri.of_string in
            let final_uri =
              Uri.add_query_params base_uri params |> Uri.to_string
            in
            Request.(get final_uri |> with_string_body "")
          in

          let test_missing_params () =
            let resp, body =
              process_request (module D) (authenticate_req "invalid_uid" [])
            in
            (check int) "endpoint does not return 400" 400
              (Response.get_status resp);
            (check (pair string string))
              "endpoint does not return valid error message"
              ( "Bad Request",
                "Required parameters are missing. Please report this error to \
                 the application requesting authentication." )
              (extract_error_msg body)
          in

          let test_invalid_app_key () =
            let resp, body =
              process_request
                (module D)
                (authenticate_req "invalid_uid"
                   [
                     ("k", [ "###not_a_valid_base64###" ]);
                     ("n", [ "doesn't matter" ]);
                     ("p", [ "doesn't matter" ]);
                     ("s", [ "doesn't matter" ]);
                   ])
            in
            (check int) "endpoint does not return 400" 400
              (Response.get_status resp);
            (check (pair string string))
              "endpoint does not return valid error message"
              ( "Bad Request",
                "Parameter is incorrectly encoded: ###not_a_valid_base64###. \
                 Please report this error to the application requesting \
                 authentication." )
              (extract_error_msg body)
          in

          let test_disapproved_app () =
            let resp, body =
              process_request
                (module D)
                (authenticate_req "invalid_uid"
                   [
                     ("k", [ "dGVzdA" ]);
                     ("n", [ "doesn't matter" ]);
                     ("p", [ "doesn't matter" ]);
                     ("s", [ "doesn't matter" ]);
                   ])
            in
            (check int) "endpoint does not return 403" 403
              (Response.get_status resp);
            (check (pair string string))
              "endpoint does not return valid error message"
              ( "Authentication Failed",
                "Application is not approved to make authentication requests."
              )
              (extract_error_msg body)
          in

          let test_non_existent_user_id () =
            let resp, _ =
              process_request
                (module D)
                (authenticate_req "invalid_uid"
                   [
                     ("k", [ Public_key.to_base64 aepk ]);
                     ("n", [ "doesn't matter" ]);
                     ("p", [ "doesn't matter" ]);
                     ("s", [ "doesn't matter" ]);
                   ])
            in
            (check int) "endpoint does not return 200" 200
              (Response.get_status resp)
          in

          let test_all_valid () =
            let aepk_str = Public_key.to_base64 aepk in
            let resp, body =
              process_request
                (module D)
                (authenticate_req uid
                   [
                     ("k", [ aepk_str ]);
                     ("n", [ "nonce" ]);
                     ("p", [ "payload" ]);
                     ("s", [ "signature" ]);
                   ])
            in
            (check int) "endpoint does not return 200" 200
              (Response.get_status resp);
            (check (option string))
              "endpoint returned invalid Content-Type"
              (Some "text/html; charset=utf-8")
              (Response.find_header resp ~key:"Content-Type");

            let expected_msg =
              String.concat ~sep:"\n"
                [
                  app_name ^ " requires authentication";
                  "You're going to be logged in as " ^ uid;
                  "Please enter your password to proceed";
                ]
            in
            (check (list string))
              "endpoint returned incorrect page"
              [ aepk_str; "nonce"; "payload"; "signature"; expected_msg ]
              (let open Soup in
              let page = parse body in
              let k_value = page $ "input[name=\"k\"]" |> R.attribute "value" in
              let n_value = page $ "input[name=\"n\"]" |> R.attribute "value" in
              let p_value = page $ "input[name=\"p\"]" |> R.attribute "value" in
              let s_value = page $ "input[name=\"s\"]" |> R.attribute "value" in
              let msg =
                page $$ "div.ok-container > p" |> to_list
                |> List.map ~f:(fun n -> texts n |> String.concat)
                |> String.concat ~sep:"\n"
              in
              [ k_value; n_value; p_value; s_value; msg ])
          in

          test_missing_params ();
          test_invalid_app_key ();
          test_non_existent_user_id ();
          test_disapproved_app ();
          test_all_valid ()))

let test_post_v1_uid_authenticate =
  test_case "POST /v1/UID/authenticate" `Quick (fun () ->
      with_tmp_file (fun db_path ->
          let module D = Daemon.Make (struct
            let db_path = db_path

            let log = ignore
          end) in
          let user_pw = "password" in
          let aepk, aesk = generate_keypair Key_purpose.Encryption in
          let aspk, assk = generate_keypair Key_purpose.Signing in
          let user =
            with_db db_path (fun db ->
                Migration.apply db ~migrations:Schema.migrations;
                let user =
                  User.create db ~id:"testuser" ~password:user_pw
                    ~name:"Test User"
                in
                ignore
                  (Application.create db ~encryption_key:aepk ~signing_key:aspk
                     ~name:"Test App");
                user)
          in

          let open Okweb in
          let authenticate_req uid body =
            Request.(
              post (sprintf "/v1/%s/authenticate" uid)
              |> with_string_body (Uri.encoded_of_query body))
          in

          let test_missing_params () =
            let resp, body =
              process_request (module D) (authenticate_req "invalid_uid" [])
            in
            (check int) "endpoint does not return 400" 400
              (Response.get_status resp);
            (check (pair string string))
              "endpoint does not return valid error message"
              ("Bad Request", "Required parameters are missing.")
              (extract_error_msg body)
          in

          let test_invalid_params () =
            let param_names = [ "k"; "n"; "p"; "s"; "password" ] in
            param_names
            |> List.filter ~f:(fun p -> String.(p <> "password"))
            |> List.iter ~f:(fun param ->
                   let invalid_base64_str = "###not_a_valid_base64###" in
                   let params =
                     param_names
                     |> List.map ~f:(function
                          | p when String.equal p param ->
                              (p, [ invalid_base64_str ])
                          | p -> (p, [ "" ]))
                   in
                   let resp, body =
                     process_request
                       (module D)
                       (authenticate_req "invalid_uid" params)
                   in
                   (check int) "endpoint does not return 400" 400
                     (Response.get_status resp);
                   (check (pair string string))
                     "endpoint does not return valid error message"
                     ( "Bad Request",
                       sprintf "Parameter is incorrectly encoded: %s"
                         invalid_base64_str )
                     (extract_error_msg body))
          in

          let test_disapproved_app () =
            let resp, body =
              process_request
                (module D)
                (authenticate_req "invalid_uid"
                   [
                     ("k", [ "dGVzdA" ]);
                     ("n", [ "" ]);
                     ("p", [ "" ]);
                     ("s", [ "" ]);
                     ("password", [ "" ]);
                   ])
            in
            (check int) "endpoint does not return 403" 403
              (Response.get_status resp);
            (check (pair string string))
              "endpoint does not return valid error message"
              ( "Authentication Failed",
                "Application is not approved to make authentication requests."
              )
              (extract_error_msg body)
          in

          let test_non_existent_user_id () =
            let resp, body =
              process_request
                (module D)
                (authenticate_req "invalid_uid"
                   [
                     ("k", [ Public_key.to_base64 aepk ]);
                     ("n", [ "" ]);
                     ("p", [ "" ]);
                     ("s", [ "" ]);
                     ("password", [ "" ]);
                   ])
            in
            (check int) "endpoint does not return 403" 403
              (Response.get_status resp);
            (check (pair string string))
              "endpoint does not return valid error message"
              ("Authentication Failed", "User ID or password are invalid.")
              (extract_error_msg body)
          in

          let test_invalid_password () =
            let resp, body =
              process_request
                (module D)
                (authenticate_req (User.get_id user)
                   [
                     ("k", [ Public_key.to_base64 aepk ]);
                     ("n", [ "" ]);
                     ("p", [ "" ]);
                     ("s", [ "" ]);
                     ("password", [ "invalid_password" ]);
                   ])
            in
            (check int) "endpoint does not return 403" 403
              (Response.get_status resp);
            (check (pair string string))
              "endpoint does not return valid error message"
              ("Authentication Failed", "User ID or password are invalid.")
              (extract_error_msg body)
          in

          let test_invalid_signature () =
            let resp, body =
              process_request
                (module D)
                (authenticate_req (User.get_id user)
                   [
                     ("k", [ Public_key.to_base64 aepk ]);
                     ("n", [ "" ]);
                     ("p", [ "payload" ]);
                     ("s", [ "invalidsignature" ]);
                     ("password", [ user_pw ]);
                   ])
            in
            (check int) "endpoint does not return 400" 400
              (Response.get_status resp);
            (check (pair string string))
              "endpoint does not return valid error message"
              ( "Verification Failed",
                "Request verification failed. Please report this error to the \
                 application requesting authentication." )
              (extract_error_msg body)
          in

          let test_all_valid () =
            let app_url = "https://some.app/random-url" in
            let nonce = generate_nonce () in
            let payload =
              encrypt
                (Plain_text.of_string app_url)
                (User.get_public_key user) aesk nonce
            in
            let signature =
              sign (Cipher_text.to_string payload ^ Nonce.to_string nonce) assk
            in
            let resp, _ =
              process_request
                (module D)
                (authenticate_req (User.get_id user)
                   [
                     ("k", [ Public_key.to_base64 aepk ]);
                     ("n", [ Nonce.to_base64 nonce ]);
                     ("p", [ Cipher_text.to_base64 payload ]);
                     ("s", [ Signature.to_base64 signature ]);
                     ("password", [ user_pw ]);
                   ])
            in
            (check int) "endpoint does not return 303" 303
              (Response.get_status resp);
            (check (option string))
              "endpoint does not return valid Location header" (Some app_url)
              (Response.find_header resp ~key:"Location")
          in

          test_missing_params ();
          test_invalid_params ();
          test_non_existent_user_id ();
          test_disapproved_app ();
          test_invalid_password ();
          test_invalid_signature ();
          test_all_valid ()))

let cases =
  [
    test_get_v1_uid_public_key;
    test_get_v1_uid_authenticate;
    test_post_v1_uid_authenticate;
  ]
