open Alcotest
open Base
open Oksql
open Okauth_server
open Okcrypto
open Utils

let test_should_create_user =
  test_case
    "should generate valid public and secret keys and insert user into database"
    `Slow (fun () ->
      with_okauth_db_in_mem (fun db ->
          let id, password, name, message =
            ("id", "password", "user name", "message")
          in
          let returned_user = User.create db ~id ~password ~name in
          let user_in_db = Option.value_exn (User.find_by_id db ~id) in
          (check bool) "inserted user doesn't equal to the returned one" true
            (User.equal returned_user user_in_db);
          (check string) "ids don't match" id (User.get_id user_in_db);
          (check string) "names don't match" name (User.get_name user_in_db);
          (check bool) "passwords don't match" true
            (verify_password ~password
               ~hash:(User.get_password_hash user_in_db));

          let pk, sk = generate_keypair Key_purpose.Encryption in
          let upk, usk =
            (User.get_public_key user_in_db, User.get_secret_key user_in_db)
          in
          let nonce = generate_nonce () in
          let ciphertext =
            encrypt (Plain_text.of_string message) pk usk nonce
          in
          let plaintext =
            decrypt ciphertext upk sk nonce |> Plain_text.to_string
          in
          (check string) "test messages don't match" message plaintext))

let test_should_remove_user =
  test_case "should remove user from the database" `Quick (fun () ->
      with_okauth_db_in_mem (fun db ->
          exec_only db
            "INSERT INTO users VALUES ('foobar', 'pw', 'pk', 'sk', 'Foo Bar');";
          let users_in_db_before = User.find_all db in
          (check int) "test user was not inserted in the db" 1
            (Array.length users_in_db_before);

          User.remove db ~id:"foobar";
          let users_in_db_after = User.find_all db in
          (check int) "test user was no removed from the db" 0
            (Array.length users_in_db_after)))

let test_should_change_password =
  test_case "should change user password" `Quick (fun () ->
      with_okauth_db_in_mem (fun db ->
          let id, password, new_password, name =
            ("id", "password", "new_password", "user name")
          in
          let user_before = User.create db ~id ~password ~name in

          (check bool) "old password couldn't be verified" true
            (verify_password ~password
               ~hash:(User.get_password_hash user_before));
          (check bool) "new password is wrongly verified" false
            (verify_password ~password:new_password
               ~hash:(User.get_password_hash user_before));

          User.change_password db ~id ~new_password;
          let user_after = Option.value_exn (User.find_by_id db ~id) in

          (check bool) "new password couldn't be verified" true
            (verify_password ~password:new_password
               ~hash:(User.get_password_hash user_after));
          (check bool) "old password is wrongly verified" false
            (verify_password ~password
               ~hash:(User.get_password_hash user_after))))

let cases =
  [
    test_should_create_user;
    test_should_remove_user;
    test_should_change_password;
  ]
